@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2008, 2009, 2010, 2011, 2012 Nicira, Inc.
+ * Copyright (c) 2008, 2009, 2010, 2011, 2012, 2017 Nicira, Inc.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -17,9 +17,12 @@
 #ifndef OPENFLOW_OPENFLOW_H
 #define OPENFLOW_OPENFLOW_H 1
 
-#include "openflow/openflow-1.0.h"
-#include "openflow/openflow-1.1.h"
-#include "openflow/openflow-1.2.h"
-#include "openflow/openflow-1.3.h"
+#include <openflow/openflow-1.0.h>
+#include <openflow/openflow-1.1.h>
+#include <openflow/openflow-1.2.h>
+#include <openflow/openflow-1.3.h>
+#include <openflow/openflow-1.4.h>
+#include <openflow/openflow-1.5.h>
+#include <openflow/openflow-1.6.h>
 
 #endif /* openflow/openflow.h */
