@@ -40,7 +40,8 @@ enum OVS_PACKED_ENUM dp_packet_source {
     DPBUF_STACK,               /* Un-movable stack space or static buffer. */
     DPBUF_STUB,                /* Starts on stack, may expand into heap. */
     DPBUF_DPDK,                /* buffer data is from DPDK allocated memory.
-                                * ref to dp_packet_init_dpdk() in dp-packet.c. */
+                                * ref to dp_packet_init_dpdk() in dp-packet.c.
+                                */
 };
 
 #define DP_PACKET_CONTEXT_SIZE 64
@@ -61,6 +62,9 @@ struct dp_packet {
     bool rss_hash_valid;        /* Is the 'rss_hash' valid? */
 #endif
     enum dp_packet_source source;  /* Source of memory allocated as 'base'. */
+
+    /* All the following elements of this struct are copied in a single call
+     * of memcpy in dp_packet_clone_with_headroom. */
     uint8_t l2_pad_size;           /* Detected l2 padding size.
                                     * Padding is non-pullable. */
     uint16_t l2_5_ofs;             /* MPLS label stack offset, or UINT16_MAX */
@@ -110,7 +114,7 @@ void dp_packet_use(struct dp_packet *, void *, size_t);
 void dp_packet_use_stub(struct dp_packet *, void *, size_t);
 void dp_packet_use_const(struct dp_packet *, const void *, size_t);
 
-void dp_packet_init_dpdk(struct dp_packet *, size_t allocated);
+void dp_packet_init_dpdk(struct dp_packet *);
 
 void dp_packet_init(struct dp_packet *, size_t);
 void dp_packet_uninit(struct dp_packet *);
@@ -348,10 +352,23 @@ dp_packet_set_l4(struct dp_packet *b, void *l4)
     b->l4_ofs = l4 ? (char *) l4 - (char *) dp_packet_data(b) : UINT16_MAX;
 }
 
+/* Returns the size of the packet from the beginning of the L3 header to the
+ * end of the L3 payload.  Hence L2 padding is not included. */
+static inline size_t
+dp_packet_l3_size(const struct dp_packet *b)
+{
+    return OVS_LIKELY(b->l3_ofs != UINT16_MAX)
+        ? (const char *)dp_packet_tail(b) - (const char *)dp_packet_l3(b)
+        - dp_packet_l2_pad_size(b)
+        : 0;
+}
+
+/* Returns the size of the packet from the beginning of the L4 header to the
+ * end of the L4 payload.  Hence L2 padding is not included. */
 static inline size_t
 dp_packet_l4_size(const struct dp_packet *b)
 {
-    return b->l4_ofs != UINT16_MAX
+    return OVS_LIKELY(b->l4_ofs != UINT16_MAX)
         ? (const char *)dp_packet_tail(b) - (const char *)dp_packet_l4(b)
         - dp_packet_l2_pad_size(b)
         : 0;
@@ -461,7 +478,96 @@ dp_packet_set_allocated(struct dp_packet *b, uint16_t s)
 {
     b->mbuf.buf_len = s;
 }
-#else
+
+/* Returns the RSS hash of the packet 'p'.  Note that the returned value is
+ * correct only if 'dp_packet_rss_valid(p)' returns true */
+static inline uint32_t
+dp_packet_get_rss_hash(struct dp_packet *p)
+{
+    return p->mbuf.hash.rss;
+}
+
+static inline void
+dp_packet_set_rss_hash(struct dp_packet *p, uint32_t hash)
+{
+    p->mbuf.hash.rss = hash;
+    p->mbuf.ol_flags |= PKT_RX_RSS_HASH;
+}
+
+static inline bool
+dp_packet_rss_valid(struct dp_packet *p)
+{
+    return p->mbuf.ol_flags & PKT_RX_RSS_HASH;
+}
+
+static inline void
+dp_packet_rss_invalidate(struct dp_packet *p OVS_UNUSED)
+{
+}
+
+static inline void
+dp_packet_mbuf_rss_flag_reset(struct dp_packet *p)
+{
+    p->mbuf.ol_flags &= ~PKT_RX_RSS_HASH;
+}
+
+/* This initialization is needed for packets that do not come from DPDK
+ * interfaces, when vswitchd is built with --with-dpdk. */
+static inline void
+dp_packet_mbuf_init(struct dp_packet *p)
+{
+    p->mbuf.ol_flags = p->mbuf.tx_offload = p->mbuf.packet_type = 0;
+    p->mbuf.nb_segs = 1;
+    p->mbuf.next = NULL;
+}
+
+static inline bool
+dp_packet_ip_checksum_valid(struct dp_packet *p)
+{
+    return (p->mbuf.ol_flags & PKT_RX_IP_CKSUM_MASK) ==
+            PKT_RX_IP_CKSUM_GOOD;
+}
+
+static inline bool
+dp_packet_ip_checksum_bad(struct dp_packet *p)
+{
+    return (p->mbuf.ol_flags & PKT_RX_IP_CKSUM_MASK) ==
+            PKT_RX_IP_CKSUM_BAD;
+}
+
+static inline bool
+dp_packet_l4_checksum_valid(struct dp_packet *p)
+{
+    return (p->mbuf.ol_flags & PKT_RX_L4_CKSUM_MASK) ==
+            PKT_RX_L4_CKSUM_GOOD;
+}
+
+static inline bool
+dp_packet_l4_checksum_bad(struct dp_packet *p)
+{
+    return (p->mbuf.ol_flags & PKT_RX_L4_CKSUM_MASK) ==
+            PKT_RX_L4_CKSUM_BAD;
+}
+
+static inline void
+reset_dp_packet_checksum_ol_flags(struct dp_packet *p)
+{
+    p->mbuf.ol_flags &= ~(PKT_RX_L4_CKSUM_GOOD | PKT_RX_L4_CKSUM_BAD |
+                          PKT_RX_IP_CKSUM_GOOD | PKT_RX_IP_CKSUM_BAD);
+}
+
+static inline bool
+dp_packet_has_flow_mark(struct dp_packet *p, uint32_t *mark)
+{
+    if (p->mbuf.ol_flags & PKT_RX_FDIR_ID) {
+        *mark = p->mbuf.hash.fdir.hi;
+        return true;
+    }
+
+    return false;
+}
+
+#else /* DPDK_NETDEV */
 static inline void *
 dp_packet_base(const struct dp_packet *b)
 {
@@ -509,7 +615,80 @@ dp_packet_set_allocated(struct dp_packet *b, uint16_t s)
 {
     b->allocated_ = s;
 }
-#endif
+
+/* Returns the RSS hash of the packet 'p'.  Note that the returned value is
+ * correct only if 'dp_packet_rss_valid(p)' returns true */
+static inline uint32_t
+dp_packet_get_rss_hash(struct dp_packet *p)
+{
+    return p->rss_hash;
+}
+
+static inline void
+dp_packet_set_rss_hash(struct dp_packet *p, uint32_t hash)
+{
+    p->rss_hash = hash;
+    p->rss_hash_valid = true;
+}
+
+static inline bool
+dp_packet_rss_valid(struct dp_packet *p)
+{
+    return p->rss_hash_valid;
+}
+
+static inline void
+dp_packet_rss_invalidate(struct dp_packet *p)
+{
+    p->rss_hash_valid = false;
+}
+
+static inline void
+dp_packet_mbuf_rss_flag_reset(struct dp_packet *p OVS_UNUSED)
+{
+}
+
+static inline void
+dp_packet_mbuf_init(struct dp_packet *p OVS_UNUSED)
+{
+}
+
+static inline bool
+dp_packet_ip_checksum_valid(struct dp_packet *p OVS_UNUSED)
+{
+    return false;
+}
+
+static inline bool
+dp_packet_ip_checksum_bad(struct dp_packet *p OVS_UNUSED)
+{
+    return false;
+}
+
+static inline bool
+dp_packet_l4_checksum_valid(struct dp_packet *p OVS_UNUSED)
+{
+    return false;
+}
+
+static inline bool
+dp_packet_l4_checksum_bad(struct dp_packet *p OVS_UNUSED)
+{
+    return false;
+}
+
+static inline void
+reset_dp_packet_checksum_ol_flags(struct dp_packet *p OVS_UNUSED)
+{
+}
+
+static inline bool
+dp_packet_has_flow_mark(struct dp_packet *p OVS_UNUSED,
+                        uint32_t *mark OVS_UNUSED)
+{
+    return false;
+}
+#endif /* DPDK_NETDEV */
 
 static inline void
 dp_packet_reset_cutlen(struct dp_packet *b)
@@ -570,123 +749,6 @@ dp_packet_reset_packet(struct dp_packet *b, int off)
     dp_packet_reset_offsets(b);
 }
 
-/* Returns the RSS hash of the packet 'p'.  Note that the returned value is
- * correct only if 'dp_packet_rss_valid(p)' returns true */
-static inline uint32_t
-dp_packet_get_rss_hash(struct dp_packet *p)
-{
-#ifdef DPDK_NETDEV
-    return p->mbuf.hash.rss;
-#else
-    return p->rss_hash;
-#endif
-}
-
-static inline void
-dp_packet_set_rss_hash(struct dp_packet *p, uint32_t hash)
-{
-#ifdef DPDK_NETDEV
-    p->mbuf.hash.rss = hash;
-    p->mbuf.ol_flags |= PKT_RX_RSS_HASH;
-#else
-    p->rss_hash = hash;
-    p->rss_hash_valid = true;
-#endif
-}
-
-static inline bool
-dp_packet_rss_valid(struct dp_packet *p)
-{
-#ifdef DPDK_NETDEV
-    return p->mbuf.ol_flags & PKT_RX_RSS_HASH;
-#else
-    return p->rss_hash_valid;
-#endif
-}
-
-static inline void
-dp_packet_rss_invalidate(struct dp_packet *p OVS_UNUSED)
-{
-#ifndef DPDK_NETDEV
-    p->rss_hash_valid = false;
-#endif
-}
-
-static inline void
-dp_packet_mbuf_rss_flag_reset(struct dp_packet *p OVS_UNUSED)
-{
-#ifdef DPDK_NETDEV
-    p->mbuf.ol_flags &= ~PKT_RX_RSS_HASH;
-#endif
-}
-
-/* This initialization is needed for packets that do not come
- * from DPDK interfaces, when vswitchd is built with --with-dpdk.
- * The DPDK rte library will still otherwise manage the mbuf.
- * We only need to initialize the mbuf ol_flags. */
-static inline void
-dp_packet_mbuf_init(struct dp_packet *p OVS_UNUSED)
-{
-#ifdef DPDK_NETDEV
-    p->mbuf.ol_flags = 0;
-#endif
-}
-
-static inline bool
-dp_packet_ip_checksum_valid(struct dp_packet *p)
-{
-#ifdef DPDK_NETDEV
-    return (p->mbuf.ol_flags & PKT_RX_IP_CKSUM_MASK) ==
-            PKT_RX_IP_CKSUM_GOOD;
-#else
-    return 0 && p;
-#endif
-}
-
-static inline bool
-dp_packet_ip_checksum_bad(struct dp_packet *p)
-{
-#ifdef DPDK_NETDEV
-    return (p->mbuf.ol_flags & PKT_RX_IP_CKSUM_MASK) ==
-            PKT_RX_IP_CKSUM_BAD;
-#else
-    return 0 && p;
-#endif
-}
-
-static inline bool
-dp_packet_l4_checksum_valid(struct dp_packet *p)
-{
-#ifdef DPDK_NETDEV
-    return (p->mbuf.ol_flags & PKT_RX_L4_CKSUM_MASK) ==
-            PKT_RX_L4_CKSUM_GOOD;
-#else
-    return 0 && p;
-#endif
-}
-
-static inline bool
-dp_packet_l4_checksum_bad(struct dp_packet *p)
-{
-#ifdef DPDK_NETDEV
-    return (p->mbuf.ol_flags & PKT_RX_L4_CKSUM_MASK) ==
-            PKT_RX_L4_CKSUM_BAD;
-#else
-    return 0 && p;
-#endif
-}
-
-#ifdef DPDK_NETDEV
-static inline void
-reset_dp_packet_checksum_ol_flags(struct dp_packet *p)
-{
-    p->mbuf.ol_flags &= ~(PKT_RX_L4_CKSUM_GOOD | PKT_RX_L4_CKSUM_BAD |
-                          PKT_RX_IP_CKSUM_GOOD | PKT_RX_IP_CKSUM_BAD);
-}
-#else
-#define reset_dp_packet_checksum_ol_flags(arg)
-#endif
-
 enum { NETDEV_MAX_BURST = 32 }; /* Maximum number packets in a batch. */
 
 struct dp_packet_batch {
@@ -755,9 +817,9 @@ dp_packet_batch_is_empty(const struct dp_packet_batch *batch)
     return !dp_packet_batch_size(batch);
 }
 
-#define DP_PACKET_BATCH_FOR_EACH(PACKET, BATCH)    \
-    for (size_t i = 0; i < dp_packet_batch_size(BATCH); i++)  \
-        if (PACKET = BATCH->packets[i], true)
+#define DP_PACKET_BATCH_FOR_EACH(IDX, PACKET, BATCH)                \
+    for (size_t IDX = 0; IDX < dp_packet_batch_size(BATCH); IDX++)  \
+        if (PACKET = BATCH->packets[IDX], true)
 
 /* Use this macro for cases where some packets in the 'BATCH' may be
  * dropped after going through each packet in the 'BATCH'.
@@ -781,19 +843,19 @@ dp_packet_batch_clone(struct dp_packet_batch *dst,
     struct dp_packet *packet;
 
     dp_packet_batch_init(dst);
-    DP_PACKET_BATCH_FOR_EACH (packet, src) {
+    DP_PACKET_BATCH_FOR_EACH (i, packet, src) {
         dp_packet_batch_add(dst, dp_packet_clone(packet));
     }
     dst->trunc = src->trunc;
 }
 
 static inline void
-dp_packet_delete_batch(struct dp_packet_batch *batch, bool may_steal)
+dp_packet_delete_batch(struct dp_packet_batch *batch, bool should_steal)
 {
-    if (may_steal) {
+    if (should_steal) {
         struct dp_packet *packet;
 
-        DP_PACKET_BATCH_FOR_EACH (packet, batch) {
+        DP_PACKET_BATCH_FOR_EACH (i, packet, batch) {
             dp_packet_delete(packet);
         }
         dp_packet_batch_init(batch);
@@ -805,7 +867,7 @@ dp_packet_batch_init_packet_fields(struct dp_packet_batch *batch)
 {
     struct dp_packet *packet;
 
-    DP_PACKET_BATCH_FOR_EACH (packet, batch) {
+    DP_PACKET_BATCH_FOR_EACH (i, packet, batch) {
         dp_packet_reset_cutlen(packet);
         packet->packet_type = htonl(PT_ETH);
     }
@@ -817,7 +879,7 @@ dp_packet_batch_apply_cutlen(struct dp_packet_batch *batch)
     if (batch->trunc) {
         struct dp_packet *packet;
 
-        DP_PACKET_BATCH_FOR_EACH (packet, batch) {
+        DP_PACKET_BATCH_FOR_EACH (i, packet, batch) {
             dp_packet_set_size(packet, dp_packet_get_send_len(packet));
             dp_packet_reset_cutlen(packet);
         }
@@ -831,7 +893,7 @@ dp_packet_batch_reset_cutlen(struct dp_packet_batch *batch)
     if (batch->trunc) {
         struct dp_packet *packet;
 
-        DP_PACKET_BATCH_FOR_EACH (packet, batch) {
+        DP_PACKET_BATCH_FOR_EACH (i, packet, batch) {
             dp_packet_reset_cutlen(packet);
         }
         batch->trunc = false;
