@@ -17,17 +17,16 @@
 #include <config.h>
 
 #include "multipath.h"
-#include <arpa/inet.h>
-#include <inttypes.h>
 #include <sys/types.h>
 #include <netinet/in.h>
+#include <arpa/inet.h>
+#include <inttypes.h>
 #include "colors.h"
 #include "nx-match.h"
 #include "openflow/nicira-ext.h"
 #include "openvswitch/dynamic-string.h"
 #include "openvswitch/ofp-actions.h"
 #include "openvswitch/ofp-errors.h"
-#include "openvswitch/ofp-util.h"
 #include "packets.h"
 #include "util.h"
 
@@ -137,7 +136,7 @@ multipath_algorithm(uint32_t hash, enum nx_mp_algorithm algorithm,
 }
 
 /* Parses 's_' as a set of arguments to the "multipath" action and initializes
- * 'mp' accordingly.  ovs-ofctl(8) describes the format parsed.
+ * 'mp' accordingly.  ovs-actions(7) describes the format parsed.
  *
  * Returns NULL if successful, otherwise a malloc()'d string describing the
  * error.  The caller is responsible for freeing the returned string.*/
@@ -172,6 +171,8 @@ multipath_parse__(struct ofpact_multipath *mp, const char *s_, char *s)
         mp->fields = NX_HASH_FIELDS_NW_SRC;
     } else if (!strcasecmp(fields, "nw_dst")) {
         mp->fields = NX_HASH_FIELDS_NW_DST;
+    } else if (!strcasecmp(fields, "symmetric_l3")) {
+        mp->fields = NX_HASH_FIELDS_SYMMETRIC_L3;
     } else {
         return xasprintf("%s: unknown fields `%s'", s_, fields);
     }
@@ -213,7 +214,7 @@ multipath_parse__(struct ofpact_multipath *mp, const char *s_, char *s)
 }
 
 /* Parses 's_' as a set of arguments to the "multipath" action and initializes
- * 'mp' accordingly.  ovs-ofctl(8) describes the format parsed.
+ * 'mp' accordingly.  ovs-actions(7) describes the format parsed.
  *
  * Returns NULL if successful, otherwise a malloc()'d string describing the
  * error.  The caller is responsible for freeing the returned string. */
@@ -226,7 +227,7 @@ multipath_parse(struct ofpact_multipath *mp, const char *s_)
     return error;
 }
 
-/* Appends a description of 'mp' to 's', in the format that ovs-ofctl(8)
+/* Appends a description of 'mp' to 's', in the format that ovs-actions(7)
  * describes. */
 void
 multipath_format(const struct ofpact_multipath *mp, struct ds *s)
