@@ -21,14 +21,8 @@
 #include "conntrack.h"
 #include "dpif.h"
 #include "ofproto-dpif-xlate.h"
-#include "openvswitch/ofp-parse.h"
 #include "unixctl.h"
 
-static void ofproto_trace(struct ofproto_dpif *, const struct flow *,
-                          const struct dp_packet *packet,
-                          const struct ofpact[], size_t ofpacts_len,
-                          struct ovs_list *next_ct_states,
-                          struct ds *);
 static void oftrace_node_destroy(struct oftrace_node *);
 
 /* Creates a new oftrace_node, populates it with the given 'type' and a copy of
@@ -55,6 +49,7 @@ oftrace_node_type_is_terminal(enum oftrace_node_type type)
     case OFT_DETAIL:
     case OFT_WARN:
     case OFT_ERROR:
+    case OFT_BUCKET:
         return true;
 
     case OFT_BRIDGE:
@@ -91,7 +86,8 @@ oftrace_node_destroy(struct oftrace_node *node)
 bool
 oftrace_add_recirc_node(struct ovs_list *recirc_queue,
                         enum oftrace_recirc_type type, const struct flow *flow,
-                        const struct dp_packet *packet, uint32_t recirc_id)
+                        const struct dp_packet *packet, uint32_t recirc_id,
+                        const uint16_t zone)
 {
     if (!recirc_id_node_find_and_ref(recirc_id)) {
         return false;
@@ -104,6 +100,7 @@ oftrace_add_recirc_node(struct ovs_list *recirc_queue,
     node->recirc_id = recirc_id;
     node->flow = *flow;
     node->flow.recirc_id = recirc_id;
+    node->flow.ct_zone = zone;
     node->packet = packet ? dp_packet_clone(packet) : NULL;
 
     return true;
@@ -171,6 +168,7 @@ oftrace_node_print_details(struct ds *output,
             ds_put_char(output, '\n');
             break;
         case OFT_TABLE:
+        case OFT_BUCKET:
         case OFT_THAW:
         case OFT_ACTION:
             ds_put_format(output, "%s\n", sub->text);
@@ -181,41 +179,11 @@ oftrace_node_print_details(struct ds *output,
     }
 }
 
-static char * OVS_WARN_UNUSED_RESULT
-parse_oftrace_options(int argc, const char *argv[],
-                      struct ovs_list *next_ct_states)
-{
-    int k;
-    struct ds ds = DS_EMPTY_INITIALIZER;
-
-    for (k = 0; k < argc; k++) {
-        if (!strncmp(argv[k], "--ct-next", 9)) {
-            if (k + 1 > argc) {
-                return xasprintf("Missing argument for option %s", argv[k]);
-            }
-
-            uint32_t ct_state;
-            if (!parse_ct_state(argv[++k], 0, &ct_state, &ds)) {
-                return ds_steal_cstr(&ds);
-            }
-            if (!validate_ct_state(ct_state, &ds)) {
-                return ds_steal_cstr(&ds);
-            }
-            oftrace_push_ct_state(next_ct_states, ct_state);
-        } else {
-            return xasprintf("Invalid option %s", argv[k]);
-        }
-    }
-
-    ds_destroy(&ds);
-    return NULL;
-}
-
 /* Parses the 'argc' elements of 'argv', ignoring argv[0].  The following
  * forms are supported:
  *
- *     - [dpname] odp_flow [OPTIONS] [-generate | packet]
- *     - bridge br_flow [OPTIONS] [-generate | packet]
+ *     - [options] [dpname] odp_flow [packet]
+ *     - [options] bridge br_flow [packet]
  *
  * On success, initializes '*ofprotop' and 'flow' and returns NULL.  On failure
  * returns a nonnull malloced error message. */
@@ -223,67 +191,140 @@ static char * OVS_WARN_UNUSED_RESULT
 parse_flow_and_packet(int argc, const char *argv[],
                       struct ofproto_dpif **ofprotop, struct flow *flow,
                       struct dp_packet **packetp,
-                      struct ovs_list *next_ct_states)
+                      struct ovs_list *next_ct_states,
+                      bool *consistent)
 {
     const struct dpif_backer *backer = NULL;
     const char *error = NULL;
     char *m_err = NULL;
     struct simap port_names = SIMAP_INITIALIZER(&port_names);
-    struct dp_packet *packet;
+    struct dp_packet *packet = NULL;
+    uint8_t *l7 = NULL;
+    size_t l7_len = 64;
     struct ofpbuf odp_key;
     struct ofpbuf odp_mask;
-    int first_option;
 
     ofpbuf_init(&odp_key, 0);
     ofpbuf_init(&odp_mask, 0);
 
-    /* Handle "-generate" or a hex string as the last argument. */
-    if (!strcmp(argv[argc - 1], "-generate")) {
-        packet = dp_packet_new(0);
-        argc--;
-    } else {
-        error = eth_from_hex(argv[argc - 1], &packet);
+    const char *args[3];
+    int n_args = 0;
+    bool generate_packet = false;
+    if (consistent) {
+        *consistent = false;
+    }
+    for (int i = 1; i < argc; i++) {
+        const char *arg = argv[i];
+        if (!strcmp(arg, "-generate") || !strcmp(arg, "--generate")) {
+            generate_packet = true;
+        } else if (!strcmp(arg, "--l7")) {
+            if (i + 1 >= argc) {
+                m_err = xasprintf("Missing argument for option %s", arg);
+                goto exit;
+            }
+
+            struct dp_packet payload;
+            memset(&payload, 0, sizeof payload);
+            dp_packet_init(&payload, 0);
+            if (dp_packet_put_hex(&payload, argv[++i], NULL)[0] != '\0') {
+                dp_packet_uninit(&payload);
+                error = "Trailing garbage in packet data";
+                goto exit;
+            }
+            free(l7);
+            l7_len = dp_packet_size(&payload);
+            l7 = dp_packet_steal_data(&payload);
+        } else if (!strcmp(arg, "--l7-len")) {
+            if (i + 1 >= argc) {
+                m_err = xasprintf("Missing argument for option %s", arg);
+                goto exit;
+            }
+            free(l7);
+            l7 = NULL;
+            l7_len = atoi(argv[++i]);
+            if (l7_len > 64000) {
+                m_err = xasprintf("%s: too much L7 data", argv[i]);
+                goto exit;
+            }
+        } else if (consistent
+                   && (!strcmp(arg, "-consistent") ||
+                       !strcmp(arg, "--consistent"))) {
+            *consistent = true;
+        } else if (!strcmp(arg, "--ct-next")) {
+            if (i + 1 >= argc) {
+                m_err = xasprintf("Missing argument for option %s", arg);
+                goto exit;
+            }
+
+            uint32_t ct_state;
+            struct ds ds = DS_EMPTY_INITIALIZER;
+            if (!parse_ct_state(argv[++i], 0, &ct_state, &ds)
+                || !validate_ct_state(ct_state, &ds)) {
+                m_err = ds_steal_cstr(&ds);
+                goto exit;
+            }
+            oftrace_push_ct_state(next_ct_states, ct_state);
+        } else if (arg[0] == '-') {
+            m_err = xasprintf("%s: unknown option", arg);
+            goto exit;
+        } else if (n_args >= ARRAY_SIZE(args)) {
+            m_err = xstrdup("too many arguments");
+            goto exit;
+        } else {
+            args[n_args++] = arg;
+        }
+    }
+
+    /* 'args' must now have one of the following forms:
+     *
+     *     odp_flow
+     *     dpname odp_flow
+     *     bridge br_flow
+     *     odp_flow packet
+     *     dpname odp_flow packet
+     *     bridge br_flow packet
+     *
+     * Parse the packet if it's there.  Note that:
+     *
+     *     - If there is one argument, there cannot be a packet.
+     *
+     *     - If there are three arguments, there must be a packet.
+     *
+     * If there is a packet, we strip it off.
+     */
+    if (!generate_packet && n_args > 1) {
+        error = eth_from_hex(args[n_args - 1], &packet);
         if (!error) {
-            argc--;
-        } else if (argc == 4) {
-            /* The 3-argument form must end in "-generate' or a hex string. */
+            n_args--;
+        } else if (n_args > 2) {
+            /* The 3-argument form must end in a hex string. */
             goto exit;
         }
         error = NULL;
     }
 
-    /* Parse options. */
-    if (argc >= 4) {
-        if (!strncmp(argv[2], "--", 2)) {
-            first_option = 2;
-        } else if (!strncmp(argv[3], "--", 2)) {
-            first_option = 3;
-        } else {
-            error = "Syntax error: invalid option";
-            goto exit;
-        }
-
-        m_err = parse_oftrace_options(argc - first_option, argv + first_option,
-                                      next_ct_states);
-        if (m_err) {
-            goto exit;
-        }
-        argc = first_option;
-    }
-
-    /* odp_flow can have its in_port specified as a name instead of port no.
-     * We do not yet know whether a given flow is a odp_flow or a br_flow.
-     * But, to know whether a flow is odp_flow through odp_flow_from_string(),
-     * we need to create a simap of name to port no. */
-    if (argc == 3) {
+    /* We stripped off the packet if there was one, so 'args' now has one of
+     * the following forms:
+     *
+     *     odp_flow
+     *     dpname odp_flow
+     *     bridge br_flow
+     *
+     * Before we parse the flow, try to identify the backer, then use that
+     * backer to assemble a collection of port names.  The port names are
+     * useful so that the user can specify ports by name instead of number in
+     * the flow. */
+    if (n_args == 2) {
+        /* args[0] might be dpname. */
         const char *dp_type;
-        if (!strncmp(argv[1], "ovs-", 4)) {
-            dp_type = argv[1] + 4;
+        if (!strncmp(args[0], "ovs-", 4)) {
+            dp_type = args[0] + 4;
         } else {
-            dp_type = argv[1];
+            dp_type = args[0];
         }
         backer = shash_find_data(&all_dpif_backers, dp_type);
-    } else if (argc == 2) {
+    } else if (n_args == 1) {
+        /* Pick default backer. */
         struct shash_node *node;
         if (shash_count(&all_dpif_backers) == 1) {
             node = shash_first(&all_dpif_backers);
@@ -306,7 +347,7 @@ parse_flow_and_packet(int argc, const char *argv[],
      * bridge is specified. If function odp_flow_key_from_string()
      * returns 0, the flow is a odp_flow. If function
      * parse_ofp_exact_flow() returns NULL, the flow is a br_flow. */
-    if (!odp_flow_from_string(argv[argc - 1], &port_names,
+    if (!odp_flow_from_string(args[n_args - 1], &port_names,
                               &odp_key, &odp_mask)) {
         if (!backer) {
             error = "Cannot find the datapath";
@@ -347,14 +388,14 @@ parse_flow_and_packet(int argc, const char *argv[],
     } else {
         char *err;
 
-        if (argc != 3) {
+        if (n_args != 2) {
             error = "Must specify bridge name";
             goto exit;
         }
 
-        *ofprotop = ofproto_dpif_lookup(argv[1]);
+        *ofprotop = ofproto_dpif_lookup_by_name(args[0]);
         if (!*ofprotop) {
-            error = "Unknown bridge name";
+            m_err = xasprintf("%s: unknown bridge", args[0]);
             goto exit;
         }
 
@@ -366,7 +407,7 @@ parse_flow_and_packet(int argc, const char *argv[],
         }
         err = parse_ofp_exact_flow(flow, NULL,
                                    ofproto_get_tun_tab(&(*ofprotop)->up),
-                                   argv[argc - 1], &map);
+                                   args[n_args - 1], &map);
         ofputil_port_map_destroy(&map);
         if (err) {
             m_err = xasprintf("Bad openflow flow syntax: %s", err);
@@ -375,16 +416,15 @@ parse_flow_and_packet(int argc, const char *argv[],
         }
     }
 
-    /* Generate a packet, if requested. */
-    if (packet) {
-        if (!dp_packet_size(packet)) {
-            flow_compose(packet, flow, 0);
-        } else {
-            /* Use the metadata from the flow and the packet argument
-             * to reconstruct the flow. */
-            pkt_metadata_from_flow(&packet->md, flow);
-            flow_extract(packet, flow);
-        }
+    if (generate_packet) {
+        /* Generate a packet, as requested. */
+        packet = dp_packet_new(0);
+        flow_compose(packet, flow, l7, l7_len);
+    } else if (packet) {
+        /* Use the metadata from the flow and the packet argument to
+         * reconstruct the flow. */
+        pkt_metadata_from_flow(&packet->md, flow);
+        flow_extract(packet, flow);
     }
 
 exit:
@@ -399,6 +439,7 @@ exit:
     ofpbuf_uninit(&odp_key);
     ofpbuf_uninit(&odp_mask);
     simap_destroy(&port_names);
+    free(l7);
     return m_err;
 }
 
@@ -421,7 +462,7 @@ ofproto_unixctl_trace(struct unixctl_conn *conn, int argc, const char *argv[],
     struct ovs_list next_ct_states = OVS_LIST_INITIALIZER(&next_ct_states);
 
     error = parse_flow_and_packet(argc, argv, &ofproto, &flow, &packet,
-                                  &next_ct_states);
+                                  &next_ct_states, NULL);
     if (!error) {
         struct ds result;
 
@@ -461,27 +502,20 @@ ofproto_unixctl_trace_actions(struct unixctl_conn *conn, int argc,
     ofpbuf_init(&ofpacts, 0);
 
     /* Parse actions. */
-    error = ofpacts_parse_actions(argv[--argc], NULL,
-                                  &ofpacts, &usable_protocols);
+    struct ofpact_parse_params pp = {
+        .port_map = NULL,
+        .ofpacts = &ofpacts,
+        .usable_protocols = &usable_protocols,
+    };
+    error = ofpacts_parse_actions(argv[--argc], &pp);
     if (error) {
         unixctl_command_reply_error(conn, error);
         free(error);
         goto exit;
     }
 
-    /* OpenFlow 1.1 and later suggest that the switch enforces certain forms of
-     * consistency between the flow and the actions.  With -consistent, we
-     * enforce consistency even for a flow supported in OpenFlow 1.0. */
-    if (!strcmp(argv[1], "-consistent")) {
-        enforce_consistency = true;
-        argv++;
-        argc--;
-    } else {
-        enforce_consistency = false;
-    }
-
     error = parse_flow_and_packet(argc, argv, &ofproto, &match.flow, &packet,
-                                  &next_ct_states);
+                                  &next_ct_states, &enforce_consistency);
     if (error) {
         unixctl_command_reply_error(conn, error);
         free(error);
@@ -504,16 +538,16 @@ ofproto_unixctl_trace_actions(struct unixctl_conn *conn, int argc,
         unixctl_command_reply_error(conn, "invalid in_port");
         goto exit;
     }
-    if (enforce_consistency) {
-        retval = ofpacts_check_consistency(ofpacts.data, ofpacts.size, &match,
-                                           u16_to_ofp(ofproto->up.max_ports),
-                                           0, ofproto->up.n_tables,
-                                           usable_protocols);
-    } else {
-        retval = ofpacts_check(ofpacts.data, ofpacts.size, &match,
-                               u16_to_ofp(ofproto->up.max_ports), 0,
-                               ofproto->up.n_tables, &usable_protocols);
-    }
+
+    struct ofpact_check_params cp = {
+        .match = &match,
+        .max_ports = u16_to_ofp(ofproto->up.max_ports),
+        .table_id = 0,
+        .n_tables = ofproto->up.n_tables,
+    };
+    retval = ofpacts_check_consistency(
+        ofpacts.data, ofpacts.size,
+        enforce_consistency ? usable_protocols : 0, &cp);
     if (!retval) {
         ovs_mutex_lock(&ofproto_mutex);
         retval = ofproto_check_ofpacts(&ofproto->up, ofpacts.data,
@@ -539,6 +573,76 @@ exit:
     free_ct_states(&next_ct_states);
 }
 
+static void
+explain_slow_path(enum slow_path_reason slow, struct ds *output)
+{
+    ds_put_cstr(output, "\nThis flow is handled by the userspace "
+                "slow path because it:");
+    for (; slow; slow = zero_rightmost_1bit(slow)) {
+        enum slow_path_reason bit = rightmost_1bit(slow);
+        ds_put_format(output, "\n  - %s.",
+                      slow_path_reason_to_explanation(bit));
+    }
+}
+
+/* Copies ODP actions from 'in' to 'out', dropping OVS_ACTION_ATTR_OUTPUT and
+ * OVS_ACTION_ATTR_RECIRC along the way. */
+static void
+prune_output_actions(const struct ofpbuf *in, struct ofpbuf *out)
+{
+    const struct nlattr *a;
+    unsigned int left;
+    NL_ATTR_FOR_EACH (a, left, in->data, in->size) {
+        if (a->nla_type == OVS_ACTION_ATTR_CLONE) {
+            struct ofpbuf in_nested;
+            nl_attr_get_nested(a, &in_nested);
+
+            size_t ofs = nl_msg_start_nested(out, OVS_ACTION_ATTR_CLONE);
+            prune_output_actions(&in_nested, out);
+            nl_msg_end_nested(out, ofs);
+        } else if (a->nla_type != OVS_ACTION_ATTR_OUTPUT &&
+                   a->nla_type != OVS_ACTION_ATTR_RECIRC) {
+            ofpbuf_put(out, a, NLA_ALIGN(a->nla_len));
+        }
+    }
+}
+
+/* Executes all of the datapath actions, except for any OVS_ACTION_ATTR_OUTPUT
+ * and OVS_ACTION_ATTR_RECIRC actions, in 'actions' on 'packet', which has the
+ * given 'flow', on 'dpif'.  The actions have slow path reason 'slow' (if any).
+ * Appends any error message to 'output'.
+ *
+ * With output and recirculation actions dropped, the only remaining side
+ * effects are from OVS_ACTION_ATTR_USERSPACE actions for executing actions to
+ * send a packet to an OpenFlow controller, IPFIX, NetFlow, and sFlow, etc. */
+static void
+execute_actions_except_outputs(struct dpif *dpif,
+                               const struct dp_packet *packet,
+                               const struct flow *flow,
+                               const struct ofpbuf *actions,
+                               enum slow_path_reason slow,
+                               struct ds *output)
+{
+    struct ofpbuf pruned_actions;
+    ofpbuf_init(&pruned_actions, 0);
+    prune_output_actions(actions, &pruned_actions);
+
+    struct dpif_execute execute = {
+        .actions = pruned_actions.data,
+        .actions_len = pruned_actions.size,
+        .needs_help = (slow & SLOW_ACTION) != 0,
+        .flow = flow,
+        .packet = dp_packet_clone_with_headroom(packet, 2),
+    };
+    int error = dpif_execute(dpif, &execute);
+    if (error) {
+        ds_put_format(output, "\nAction execution failed (%s)\n.",
+                      ovs_strerror(error));
+    }
+    dp_packet_delete(execute.packet);
+    ofpbuf_uninit(&pruned_actions);
+}
+
 static void
 ofproto_trace__(struct ofproto_dpif *ofproto, const struct flow *flow,
                 const struct dp_packet *packet, struct ovs_list *recirc_queue,
@@ -593,23 +697,18 @@ ofproto_trace__(struct ofproto_dpif *ofproto, const struct flow *flow,
         ds_put_format(output,
                       "\nTranslation failed (%s), packet is dropped.\n",
                       xlate_strerror(error));
-    } else if (xout.slow) {
-        enum slow_path_reason slow;
-
-        ds_put_cstr(output, "\nThis flow is handled by the userspace "
-                    "slow path because it:");
-
-        slow = xout.slow;
-        while (slow) {
-            enum slow_path_reason bit = rightmost_1bit(slow);
-
-            ds_put_format(output, "\n\t- %s.",
-                          slow_path_reason_to_explanation(bit));
-
-            slow &= ~bit;
+    } else {
+        if (xout.slow) {
+            explain_slow_path(xout.slow, output);
+        }
+        if (packet) {
+            execute_actions_except_outputs(ofproto->backer->dpif, packet,
+                                           &initial_flow, &odp_actions,
+                                           xout.slow, output);
         }
     }
 
+
     xlate_out_uninit(&xout);
     ofpbuf_uninit(&odp_actions);
     oftrace_node_list_destroy(&trace);
@@ -624,7 +723,7 @@ ofproto_trace__(struct ofproto_dpif *ofproto, const struct flow *flow,
  *
  * If 'ofpacts' is nonnull then its 'ofpacts_len' bytes specify the actions to
  * trace, otherwise the actions are determined by a flow table lookup. */
-static void
+void
 ofproto_trace(struct ofproto_dpif *ofproto, const struct flow *flow,
               const struct dp_packet *packet,
               const struct ofpact ofpacts[], size_t ofpacts_len,
@@ -641,7 +740,7 @@ ofproto_trace(struct ofproto_dpif *ofproto, const struct flow *flow,
         ds_put_format(output, "\nrecirc(%#"PRIx32")",
                       recirc_node->recirc_id);
 
-        if (recirc_node->type == OFT_RECIRC_CONNTRACK) {
+        if (next_ct_states && recirc_node->type == OFT_RECIRC_CONNTRACK) {
             uint32_t ct_state;
             if (ovs_list_is_empty(next_ct_states)) {
                 ct_state = CS_TRACKED | CS_NEW;
