@@ -18,14 +18,23 @@
 
 #include <stdbool.h>
 
-struct controller_ctx;
 struct ovsdb_idl;
+struct ovsdb_idl_txn;
 struct ovsrec_bridge;
+struct ovsrec_bridge_table;
+struct sbrec_chassis_table;
+struct sbrec_sb_global;
+struct ovsrec_open_vswitch_table;
 
 void encaps_register_ovs_idl(struct ovsdb_idl *);
-void encaps_run(struct controller_ctx *,
-                const struct ovsrec_bridge *br_int, const char *chassis_id);
-bool encaps_cleanup(struct controller_ctx *,
+void encaps_run(struct ovsdb_idl_txn *ovs_idl_txn,
+                const struct ovsrec_bridge_table *,
+                const struct ovsrec_bridge *br_int,
+                const struct sbrec_chassis_table *,
+                const char *chassis_id,
+                const struct sbrec_sb_global *);
+
+bool encaps_cleanup(struct ovsdb_idl_txn *ovs_idl_txn,
                     const struct ovsrec_bridge *br_int);
 
 #endif /* ovn/encaps.h */
