@@ -14,8 +14,9 @@
  */
 
 #include <config.h>
-#include "gchassis.h"
 #include "lflow.h"
+#include "coverage.h"
+#include "gchassis.h"
 #include "lport.h"
 #include "ofctrl.h"
 #include "openvswitch/dynamic-string.h"
@@ -25,14 +26,17 @@
 #include "ovn-controller.h"
 #include "ovn/actions.h"
 #include "ovn/expr.h"
-#include "ovn/lib/ovn-dhcp.h"
+#include "ovn/lib/ovn-l7.h"
 #include "ovn/lib/ovn-sb-idl.h"
+#include "ovn/lib/extend-table.h"
 #include "packets.h"
 #include "physical.h"
 #include "simap.h"
 #include "sset.h"
 
 VLOG_DEFINE_THIS_MODULE(lflow);
+
+COVERAGE_DEFINE(lflow_run);
 
 /* Symbol table. */
 
@@ -46,30 +50,36 @@ lflow_init(void)
 }
 
 struct lookup_port_aux {
-    struct ovsdb_idl *ovnsb_idl;
+    struct ovsdb_idl_index *sbrec_multicast_group_by_name_datapath;
+    struct ovsdb_idl_index *sbrec_port_binding_by_name;
     const struct sbrec_datapath_binding *dp;
 };
 
 struct condition_aux {
-    struct ovsdb_idl *ovnsb_idl;
+    struct ovsdb_idl_index *sbrec_chassis_by_name;
+    struct ovsdb_idl_index *sbrec_port_binding_by_name;
     const struct sbrec_chassis *chassis;
     const struct sset *active_tunnels;
-    const struct chassis_index *chassis_index;
 };
 
-static void consider_logical_flow(struct controller_ctx *ctx,
-                                  const struct chassis_index *chassis_index,
-                                  const struct sbrec_logical_flow *lflow,
-                                  const struct hmap *local_datapaths,
-                                  struct group_table *group_table,
-                                  const struct sbrec_chassis *chassis,
-                                  struct hmap *dhcp_opts,
-                                  struct hmap *dhcpv6_opts,
-                                  uint32_t *conj_id_ofs,
-                                  const struct shash *addr_sets,
-                                  struct hmap *flow_table,
-                                  struct sset *active_tunnels,
-                                  struct sset *local_lport_ids);
+static void consider_logical_flow(
+    struct ovsdb_idl_index *sbrec_chassis_by_name,
+    struct ovsdb_idl_index *sbrec_multicast_group_by_name_datapath,
+    struct ovsdb_idl_index *sbrec_port_binding_by_name,
+    const struct sbrec_logical_flow *,
+    const struct hmap *local_datapaths,
+    const struct sbrec_chassis *,
+    struct hmap *dhcp_opts,
+    struct hmap *dhcpv6_opts,
+    struct hmap *nd_ra_opts,
+    const struct shash *addr_sets,
+    const struct shash *port_groups,
+    const struct sset *active_tunnels,
+    const struct sset *local_lport_ids,
+    uint32_t *conj_id_ofs,
+    struct hmap *flow_table,
+    struct ovn_extend_table *group_table,
+    struct ovn_extend_table *meter_table);
 
 static bool
 lookup_port_cb(const void *aux_, const char *port_name, unsigned int *portp)
@@ -77,14 +87,14 @@ lookup_port_cb(const void *aux_, const char *port_name, unsigned int *portp)
     const struct lookup_port_aux *aux = aux_;
 
     const struct sbrec_port_binding *pb
-        = lport_lookup_by_name(aux->ovnsb_idl, port_name);
+        = lport_lookup_by_name(aux->sbrec_port_binding_by_name, port_name);
     if (pb && pb->datapath == aux->dp) {
         *portp = pb->tunnel_key;
         return true;
     }
 
-    const struct sbrec_multicast_group *mg
-        = mcgroup_lookup_by_dp_name(aux->ovnsb_idl, aux->dp, port_name);
+    const struct sbrec_multicast_group *mg = mcgroup_lookup_by_dp_name(
+        aux->sbrec_multicast_group_by_name_datapath, aux->dp, port_name);
     if (mg) {
         *portp = mg->tunnel_key;
         return true;
@@ -99,7 +109,7 @@ is_chassis_resident_cb(const void *c_aux_, const char *port_name)
     const struct condition_aux *c_aux = c_aux_;
 
     const struct sbrec_port_binding *pb
-        = lport_lookup_by_name(c_aux->ovnsb_idl, port_name);
+        = lport_lookup_by_name(c_aux->sbrec_port_binding_by_name, port_name);
     if (!pb) {
         return false;
     }
@@ -108,8 +118,8 @@ is_chassis_resident_cb(const void *c_aux_, const char *port_name)
         return pb->chassis && pb->chassis == c_aux->chassis;
     } else {
         struct ovs_list *gateway_chassis;
-        gateway_chassis = gateway_chassis_get_ordered(pb,
-                                                      c_aux->chassis_index);
+        gateway_chassis = gateway_chassis_get_ordered(
+            c_aux->sbrec_chassis_by_name, pb);
         if (gateway_chassis) {
             bool active = gateway_chassis_is_active(gateway_chassis,
                                                     c_aux->chassis,
@@ -128,26 +138,24 @@ is_switch(const struct sbrec_datapath_binding *ldp)
 
 }
 
-static bool
-is_gateway_router(const struct sbrec_datapath_binding *ldp,
-                  const struct hmap *local_datapaths)
-{
-    struct local_datapath *ld =
-        get_local_datapath(local_datapaths, ldp->tunnel_key);
-    return ld ? ld->has_local_l3gateway : false;
-}
-
 /* Adds the logical flows from the Logical_Flow table to flow tables. */
 static void
-add_logical_flows(struct controller_ctx *ctx,
-                  const struct chassis_index *chassis_index,
-                  const struct hmap *local_datapaths,
-                  struct group_table *group_table,
-                  const struct sbrec_chassis *chassis,
-                  const struct shash *addr_sets,
-                  struct hmap *flow_table,
-                  struct sset *active_tunnels,
-                  struct sset *local_lport_ids)
+add_logical_flows(
+    struct ovsdb_idl_index *sbrec_chassis_by_name,
+    struct ovsdb_idl_index *sbrec_multicast_group_by_name_datapath,
+    struct ovsdb_idl_index *sbrec_port_binding_by_name,
+    const struct sbrec_dhcp_options_table *dhcp_options_table,
+    const struct sbrec_dhcpv6_options_table *dhcpv6_options_table,
+    const struct sbrec_logical_flow_table *logical_flow_table,
+    const struct hmap *local_datapaths,
+    const struct sbrec_chassis *chassis,
+    const struct shash *addr_sets,
+    const struct shash *port_groups,
+    const struct sset *active_tunnels,
+    const struct sset *local_lport_ids,
+    struct hmap *flow_table,
+    struct ovn_extend_table *group_table,
+    struct ovn_extend_table *meter_table)
 {
     uint32_t conj_id_ofs = 1;
     const struct sbrec_logical_flow *lflow;
@@ -155,45 +163,57 @@ add_logical_flows(struct controller_ctx *ctx,
     struct hmap dhcp_opts = HMAP_INITIALIZER(&dhcp_opts);
     struct hmap dhcpv6_opts = HMAP_INITIALIZER(&dhcpv6_opts);
     const struct sbrec_dhcp_options *dhcp_opt_row;
-    SBREC_DHCP_OPTIONS_FOR_EACH(dhcp_opt_row, ctx->ovnsb_idl) {
+    SBREC_DHCP_OPTIONS_TABLE_FOR_EACH (dhcp_opt_row, dhcp_options_table) {
         dhcp_opt_add(&dhcp_opts, dhcp_opt_row->name, dhcp_opt_row->code,
                      dhcp_opt_row->type);
     }
 
 
     const struct sbrec_dhcpv6_options *dhcpv6_opt_row;
-    SBREC_DHCPV6_OPTIONS_FOR_EACH(dhcpv6_opt_row, ctx->ovnsb_idl) {
+    SBREC_DHCPV6_OPTIONS_TABLE_FOR_EACH (dhcpv6_opt_row,
+                                         dhcpv6_options_table) {
        dhcp_opt_add(&dhcpv6_opts, dhcpv6_opt_row->name, dhcpv6_opt_row->code,
                     dhcpv6_opt_row->type);
     }
 
-    SBREC_LOGICAL_FLOW_FOR_EACH (lflow, ctx->ovnsb_idl) {
-        consider_logical_flow(ctx, chassis_index,
+    struct hmap nd_ra_opts = HMAP_INITIALIZER(&nd_ra_opts);
+    nd_ra_opts_init(&nd_ra_opts);
+
+    SBREC_LOGICAL_FLOW_TABLE_FOR_EACH (lflow, logical_flow_table) {
+        consider_logical_flow(sbrec_chassis_by_name,
+                              sbrec_multicast_group_by_name_datapath,
+                              sbrec_port_binding_by_name,
                               lflow, local_datapaths,
-                              group_table, chassis,
-                              &dhcp_opts, &dhcpv6_opts, &conj_id_ofs,
-                              addr_sets, flow_table, active_tunnels,
-                              local_lport_ids);
+                              chassis, &dhcp_opts, &dhcpv6_opts, &nd_ra_opts,
+                              addr_sets, port_groups, active_tunnels,
+                              local_lport_ids, &conj_id_ofs,
+                              flow_table, group_table, meter_table);
     }
 
     dhcp_opts_destroy(&dhcp_opts);
     dhcp_opts_destroy(&dhcpv6_opts);
+    nd_ra_opts_destroy(&nd_ra_opts);
 }
 
 static void
-consider_logical_flow(struct controller_ctx *ctx,
-                      const struct chassis_index *chassis_index,
-                      const struct sbrec_logical_flow *lflow,
-                      const struct hmap *local_datapaths,
-                      struct group_table *group_table,
-                      const struct sbrec_chassis *chassis,
-                      struct hmap *dhcp_opts,
-                      struct hmap *dhcpv6_opts,
-                      uint32_t *conj_id_ofs,
-                      const struct shash *addr_sets,
-                      struct hmap *flow_table,
-                      struct sset *active_tunnels,
-                      struct sset *local_lport_ids)
+consider_logical_flow(
+    struct ovsdb_idl_index *sbrec_chassis_by_name,
+    struct ovsdb_idl_index *sbrec_multicast_group_by_name_datapath,
+    struct ovsdb_idl_index *sbrec_port_binding_by_name,
+    const struct sbrec_logical_flow *lflow,
+    const struct hmap *local_datapaths,
+    const struct sbrec_chassis *chassis,
+    struct hmap *dhcp_opts,
+    struct hmap *dhcpv6_opts,
+    struct hmap *nd_ra_opts,
+    const struct shash *addr_sets,
+    const struct shash *port_groups,
+    const struct sset *active_tunnels,
+    const struct sset *local_lport_ids,
+    uint32_t *conj_id_ofs,
+    struct hmap *flow_table,
+    struct ovn_extend_table *group_table,
+    struct ovn_extend_table *meter_table)
 {
     /* Determine translation of logical table IDs to physical table IDs. */
     bool ingress = !strcmp(lflow->pipeline, "ingress");
@@ -224,6 +244,7 @@ consider_logical_flow(struct controller_ctx *ctx,
         .symtab = &symtab,
         .dhcp_opts = dhcp_opts,
         .dhcpv6_opts = dhcpv6_opts,
+        .nd_ra_opts = nd_ra_opts,
 
         .pipeline = ingress ? OVNACT_P_INGRESS : OVNACT_P_EGRESS,
         .n_tables = LOG_PIPELINE_LEN,
@@ -243,35 +264,12 @@ consider_logical_flow(struct controller_ctx *ctx,
         return;
     }
 
-    /* Encode OVN logical actions into OpenFlow. */
-    uint64_t ofpacts_stub[1024 / 8];
-    struct ofpbuf ofpacts = OFPBUF_STUB_INITIALIZER(ofpacts_stub);
-    struct lookup_port_aux aux = {
-        .ovnsb_idl = ctx->ovnsb_idl,
-        .dp = lflow->logical_datapath
-    };
-    struct ovnact_encode_params ep = {
-        .lookup_port = lookup_port_cb,
-        .aux = &aux,
-        .is_switch = is_switch(ldp),
-        .is_gateway_router = is_gateway_router(ldp, local_datapaths),
-        .group_table = group_table,
-
-        .pipeline = ingress ? OVNACT_P_INGRESS : OVNACT_P_EGRESS,
-        .ingress_ptable = OFTABLE_LOG_INGRESS_PIPELINE,
-        .egress_ptable = OFTABLE_LOG_EGRESS_PIPELINE,
-        .output_ptable = output_ptable,
-        .mac_bind_ptable = OFTABLE_MAC_BINDING,
-    };
-    ovnacts_encode(ovnacts.data, ovnacts.size, &ep, &ofpacts);
-    ovnacts_free(ovnacts.data, ovnacts.size);
-    ofpbuf_uninit(&ovnacts);
-
     /* Translate OVN match into table of OpenFlow matches. */
     struct hmap matches;
     struct expr *expr;
 
-    expr = expr_parse_string(lflow->match, &symtab, addr_sets, &error);
+    expr = expr_parse_string(lflow->match, &symtab, addr_sets, port_groups,
+                             &error);
     if (!error) {
         if (prereqs) {
             expr = expr_combine(EXPR_T_AND, expr, prereqs);
@@ -284,19 +282,57 @@ consider_logical_flow(struct controller_ctx *ctx,
         VLOG_WARN_RL(&rl, "error parsing match \"%s\": %s",
                      lflow->match, error);
         expr_destroy(prereqs);
-        ofpbuf_uninit(&ofpacts);
         free(error);
+        ovnacts_free(ovnacts.data, ovnacts.size);
+        ofpbuf_uninit(&ovnacts);
         return;
     }
 
-    struct condition_aux cond_aux = { ctx->ovnsb_idl, chassis, active_tunnels,
-                                      chassis_index};
+    struct lookup_port_aux aux = {
+        .sbrec_multicast_group_by_name_datapath
+            = sbrec_multicast_group_by_name_datapath,
+        .sbrec_port_binding_by_name = sbrec_port_binding_by_name,
+        .dp = lflow->logical_datapath
+    };
+    struct condition_aux cond_aux = {
+        .sbrec_chassis_by_name = sbrec_chassis_by_name,
+        .sbrec_port_binding_by_name = sbrec_port_binding_by_name,
+        .chassis = chassis,
+        .active_tunnels = active_tunnels,
+    };
     expr = expr_simplify(expr, is_chassis_resident_cb, &cond_aux);
     expr = expr_normalize(expr);
     uint32_t n_conjs = expr_to_matches(expr, lookup_port_cb, &aux,
                                        &matches);
     expr_destroy(expr);
 
+    if (hmap_is_empty(&matches)) {
+        ovnacts_free(ovnacts.data, ovnacts.size);
+        ofpbuf_uninit(&ovnacts);
+        expr_matches_destroy(&matches);
+        return;
+    }
+
+    /* Encode OVN logical actions into OpenFlow. */
+    uint64_t ofpacts_stub[1024 / 8];
+    struct ofpbuf ofpacts = OFPBUF_STUB_INITIALIZER(ofpacts_stub);
+    struct ovnact_encode_params ep = {
+        .lookup_port = lookup_port_cb,
+        .aux = &aux,
+        .is_switch = is_switch(ldp),
+        .group_table = group_table,
+        .meter_table = meter_table,
+
+        .pipeline = ingress ? OVNACT_P_INGRESS : OVNACT_P_EGRESS,
+        .ingress_ptable = OFTABLE_LOG_INGRESS_PIPELINE,
+        .egress_ptable = OFTABLE_LOG_EGRESS_PIPELINE,
+        .output_ptable = output_ptable,
+        .mac_bind_ptable = OFTABLE_MAC_BINDING,
+    };
+    ovnacts_encode(ovnacts.data, ovnacts.size, &ep, &ofpacts);
+    ovnacts_free(ovnacts.data, ovnacts.size);
+    ofpbuf_uninit(&ovnacts);
+
     /* Prepare the OpenFlow matches for adding to the flow table. */
     struct expr_match *m;
     HMAP_FOR_EACH (m, hmap_node, &matches) {
@@ -360,12 +396,12 @@ put_load(const uint8_t *data, size_t len,
 }
 
 static void
-consider_neighbor_flow(struct controller_ctx *ctx,
+consider_neighbor_flow(struct ovsdb_idl_index *sbrec_port_binding_by_name,
                        const struct sbrec_mac_binding *b,
                        struct hmap *flow_table)
 {
     const struct sbrec_port_binding *pb
-        = lport_lookup_by_name(ctx->ovnsb_idl, b->logical_port);
+        = lport_lookup_by_name(sbrec_port_binding_by_name, b->logical_port);
     if (!pb) {
         return;
     }
@@ -411,32 +447,47 @@ consider_neighbor_flow(struct controller_ctx *ctx,
 /* Adds an OpenFlow flow to flow tables for each MAC binding in the OVN
  * southbound database. */
 static void
-add_neighbor_flows(struct controller_ctx *ctx,
+add_neighbor_flows(struct ovsdb_idl_index *sbrec_port_binding_by_name,
+                   const struct sbrec_mac_binding_table *mac_binding_table,
                    struct hmap *flow_table)
 {
     const struct sbrec_mac_binding *b;
-    SBREC_MAC_BINDING_FOR_EACH (b, ctx->ovnsb_idl) {
-        consider_neighbor_flow(ctx, b, flow_table);
+    SBREC_MAC_BINDING_TABLE_FOR_EACH (b, mac_binding_table) {
+        consider_neighbor_flow(sbrec_port_binding_by_name, b, flow_table);
     }
 }
 
 /* Translates logical flows in the Logical_Flow table in the OVN_SB database
  * into OpenFlow flows.  See ovn-architecture(7) for more information. */
 void
-lflow_run(struct controller_ctx *ctx,
+lflow_run(struct ovsdb_idl_index *sbrec_chassis_by_name,
+          struct ovsdb_idl_index *sbrec_multicast_group_by_name_datapath,
+          struct ovsdb_idl_index *sbrec_port_binding_by_name,
+          const struct sbrec_dhcp_options_table *dhcp_options_table,
+          const struct sbrec_dhcpv6_options_table *dhcpv6_options_table,
+          const struct sbrec_logical_flow_table *logical_flow_table,
+          const struct sbrec_mac_binding_table *mac_binding_table,
           const struct sbrec_chassis *chassis,
-          const struct chassis_index *chassis_index,
           const struct hmap *local_datapaths,
-          struct group_table *group_table,
           const struct shash *addr_sets,
+          const struct shash *port_groups,
+          const struct sset *active_tunnels,
+          const struct sset *local_lport_ids,
           struct hmap *flow_table,
-          struct sset *active_tunnels,
-          struct sset *local_lport_ids)
+          struct ovn_extend_table *group_table,
+          struct ovn_extend_table *meter_table)
 {
-    add_logical_flows(ctx, chassis_index, local_datapaths,
-                      group_table, chassis, addr_sets, flow_table,
-                      active_tunnels, local_lport_ids);
-    add_neighbor_flows(ctx, flow_table);
+    COVERAGE_INC(lflow_run);
+
+    add_logical_flows(sbrec_chassis_by_name,
+                      sbrec_multicast_group_by_name_datapath,
+                      sbrec_port_binding_by_name, dhcp_options_table,
+                      dhcpv6_options_table, logical_flow_table,
+                      local_datapaths, chassis, addr_sets, port_groups,
+                      active_tunnels, local_lport_ids, flow_table, group_table,
+                      meter_table);
+    add_neighbor_flows(sbrec_port_binding_by_name, mac_binding_table,
+                       flow_table);
 }
 
 void
