@@ -27,12 +27,17 @@ import ovs.ovsuuid
 import ovs.poller
 import ovs.stream
 import ovs.util
+import ovs.vlog
 from ovs.db import data
 from ovs.db import error
 from ovs.fatal_signal import signal_alarm
 
 import six
 
+vlog = ovs.vlog.Vlog("test-ovsdb")
+vlog.set_levels_from_string("console:dbg")
+vlog.init(None)
+
 
 def unbox_json(json):
     if type(json) == list and len(json) == 1:
@@ -162,7 +167,7 @@ def get_simple_printable_row_string(row, columns):
             s += "%s=%s " % (column, value)
     s = s.strip()
     s = re.sub('""|,|u?\'', "", s)
-    s = re.sub('UUID\(([^)]+)\)', r'\1', s)
+    s = re.sub(r'UUID\(([^)]+)\)', r'\1', s)
     s = re.sub('False', 'false', s)
     s = re.sub('True', 'true', s)
     s = re.sub(r'(ba)=([^[][^ ]*) ', r'\1=[\2] ', s)
@@ -240,6 +245,16 @@ def print_idl(idl, step):
             print(''.join(s))
             n += 1
 
+    if "singleton" in idl.tables:
+        sng = idl.tables["singleton"].rows
+        for row in six.itervalues(sng):
+            s = ["%03d:" % step]
+            s.append(" name=%s" % row.name)
+            if hasattr(row, "uuid"):
+                s.append(" uuid=%s" % row.uuid)
+            print(''.join(s))
+            n += 1
+
     if not n:
         print("%03d: empty" % step)
     sys.stdout.flush()
@@ -552,6 +567,11 @@ def do_idl(schema_file, remote, *commands):
     track_notify = False
 
     if remote.startswith("ssl:"):
+        if len(commands) < 3:
+            sys.stderr.write("SSL connection requires private key, "
+                             "certificate for private key, and peer CA "
+                             "certificate as arguments\n")
+            sys.exit(1)
         ovs.stream.Stream.ssl_set_private_key_file(commands[0])
         ovs.stream.Stream.ssl_set_certificate_file(commands[1])
         ovs.stream.Stream.ssl_set_ca_cert_file(commands[2])
@@ -580,9 +600,16 @@ def do_idl(schema_file, remote, *commands):
         idl.index_create("simple3", "simple3_by_name")
 
     if commands:
-        error, stream = ovs.stream.Stream.open_block(
-            ovs.stream.Stream.open(remote))
-        if error:
+        remotes = remote.split(',')
+        stream = None
+        for r in remotes:
+            error, stream = ovs.stream.Stream.open_block(
+                ovs.stream.Stream.open(r), 2000)
+            if not error and stream:
+                break
+            stream = None
+
+        if not stream:
             sys.stderr.write("failed to connect to \"%s\"" % remote)
             sys.exit(1)
         rpc = ovs.jsonrpc.Connection(stream)
@@ -657,11 +684,11 @@ def do_idl(schema_file, remote, *commands):
             request = ovs.jsonrpc.Message.create_request("transact", json)
             error, reply = rpc.transact_block(request)
             if error:
-                sys.stderr.write("jsonrpc transaction failed: %s"
+                sys.stderr.write("jsonrpc transaction failed: %s\n"
                                  % os.strerror(error))
                 sys.exit(1)
             elif reply.error is not None:
-                sys.stderr.write("jsonrpc transaction failed: %s"
+                sys.stderr.write("jsonrpc transaction failed: %s\n"
                                  % reply.error)
                 sys.exit(1)
 
@@ -710,11 +737,11 @@ def do_idl_passive(schema_file, remote, *commands):
         request = ovs.jsonrpc.Message.create_request("transact", json)
         error, reply = rpc.transact_block(request)
         if error:
-            sys.stderr.write("jsonrpc transaction failed: %s"
+            sys.stderr.write("jsonrpc transaction failed: %s\n"
                              % os.strerror(error))
             sys.exit(1)
         elif reply.error is not None:
-            sys.stderr.write("jsonrpc transaction failed: %s"
+            sys.stderr.write("jsonrpc transaction failed: %s\n"
                              % reply.error)
             sys.exit(1)
 
@@ -800,6 +827,7 @@ def main(argv):
         sys.stderr.write("%s: %s\n" % (ovs.util.PROGRAM_NAME, geo.msg))
         sys.exit(1)
 
+    timeout = None
     for key, value in options:
         if key in ['-h', '--help']:
             usage()
@@ -811,10 +839,11 @@ def main(argv):
             except TypeError:
                 raise error.Error("value %s on -t or --timeout is not at "
                                   "least 1" % value)
-            signal_alarm(timeout)
         else:
             sys.exit(0)
 
+    signal_alarm(timeout)
+
     if not args:
         sys.stderr.write("%s: missing command argument "
                          "(use --help for help)\n" % ovs.util.PROGRAM_NAME)
